import numpy as np
from ngsolve import *
from netgen.geom2d import SplineGeometry
geo = SplineGeometry()

def MakeCorridor():
    pnts =[(0,0),
           (1.3,0),
           (1.35,0.15),
           (1.4, 0.15),
           (1.7, 0.15),
           (1.75, 0.15),
           (1.8, 0.0),
           (3.0, 0.0),
           (3.0, 0.1),
           (3.0, 0.4),
           (3.0, 0.5),
           (1.8, 0.5),
           (1.75, 0.35),
           (1.7, 0.35),
           (1.4, 0.35),
           (1.35,0.35),
           (1.3, 0.5),
           (0,0.5),
           (0,0.4),
           (0,0.1)]
    
    p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19,p20 = [geo.AppendPoint(*pnt) for pnt in pnts]
    geo.Append(["line",p1,p2],leftdomain=1,rightdomain=0,bc="wall")
    geo.Append(["spline3",p2,p3,p4],leftdomain=1,rightdomain=0,bc="wall")
    geo.Append(["line",p4,p5],leftdomain=1,rightdomain=0,bc="wall")
    geo.Append(["spline3",p5,p6,p7],leftdomain=1,rightdomain=0,bc="wall")
    geo.Append(["line",p7,p8],leftdomain=1,rightdomain=0,bc="wall")
    geo.Append(["line",p8,p9],leftdomain=1,rightdomain=0,bc="wall")
    geo.Append(["line",p9,p10],leftdomain=1,rightdomain=0,bc="outflow")
    geo.Append(["line",p10,p11],leftdomain=1,rightdomain=0,bc="wall")
    geo.Append(["line",p11,p12],leftdomain=1,rightdomain=0,bc="wall")
    geo.Append(["spline3",p12,p13,p14],leftdomain=1,rightdomain=0,bc="wall")
    geo.Append(["line",p14,p15],leftdomain=1,rightdomain=0,bc="wall")
    geo.Append(["spline3",p15,p16,p17],leftdomain=1,rightdomain=0,bc="wall")
    geo.Append(["line",p17,p18],leftdomain=1,rightdomain=0,bc="wall")
    geo.Append(["line",p18,p19],leftdomain=1,rightdomain=0,bc="inflow")
    geo.Append(["line",p19,p20],leftdomain=1,rightdomain=0,bc="wall")
    geo.Append(["line",p20,p1],leftdomain=1,rightdomain=0,bc="wall")
    
    #[geo.Append(c,l,r,bc) for c,l,r,bc in curves]
    geo.SetMaterial(1,"dummy")
    return geo


def FPIMEX4(fes, u, a, m, inv, conv, tau, N):

    d = u.vec.CreateVector()    
    fbc_out = u.vec.CreateVector()
    fbc_in = u.vec.CreateVector()
    z = u.vec.CreateVector()
    w = u.vec.CreateVector()
    q = u.vec.CreateVector()
    c1 = u.vec.CreateVector()
    c2 = u.vec.CreateVector()
    c3 = u.vec.CreateVector()
    c4 = u.vec.CreateVector()
    b1 = u.vec.CreateVector()
    b2 = u.vec.CreateVector()
    b3 = u.vec.CreateVector()

    q = m.mat * u.vec

    # 4 stage IMEX scheme
    for index in np.arange(1,N+1,1):

        conv.Apply(u.vec,w)
        
        c1.data = w
        d.data = q - tau * 0.5 * c1
        z.data = inv * d
        
        conv.Apply(z,w)

        c2.data = w
        b1.data = a.mat * z
        d.data = q - tau * (11.0/18.0 * c1 + 1.0/18.0 * c2 + 1.0/6.0 * b1)
        z.data = inv * d
    
        conv.Apply(z,w)
        
        c3.data = w 
        b2.data = a.mat * z
        d.data = q - tau * (5.0/6.0 * c1 - 5.0/6.0 * c2 + 0.5 * c3 - 0.5 * b1 + 0.5 * b2)
        z.data = inv * d
        
        conv.Apply(z,w)
        
        c4.data = w
        b3.data = a.mat * z
        d.data = q - tau * (0.25 * c1 + 1.75 * c2 + 0.75 * c3 - 1.75 * c4 + 1.5 * b1 - 1.5 * b2 + 0.5 * b3)
        
        u.vec.data = inv * d
        Redraw(True)

def localsolver(u, mesh, sorted_vertices, vertex_to_elements, pointarray, costs):
    
    for p in sorted_vertices:

        for el in vertex_to_elements[p]:
        
            vert = mesh[ElementId(el)].vertices
            vertices = np.array([vert[0].nr,vert[1].nr, vert[2].nr])
        
            AB = np.extract(vertices != p, vertices)

            if u.vec[AB[0]] <= u.vec[AB[1]]:
                TA = u.vec[AB[0]]
                TB = u.vec[AB[1]]
            else:
                TB = u.vec[AB[0]]
                TA = u.vec[AB[1]]

            TC = u.vec[p]
            
            xA = pointarray[:,AB[0]]
            xB = pointarray[:,AB[1]]
            xC = pointarray[:,p]
            
            xAB = xA - xB
            xAC = xA - xC
            xBC = xB - xC
            
            a = sqrt(xBC[0]**2 + xBC[1]**2)
            b = sqrt(xAC[0]**2 + xAC[1]**2)
            c = sqrt(xAB[0]**2 + xAB[1]**2)

            beta = np.arccos(np.dot(xAB, xAC)/(b*c))
            alpha = np.arccos(np.dot(-xAB, xBC)/(a*c))
            gamma = np.arccos(np.dot(-xAC, -xBC)/(b*a))
            
            if ((TB-TA) <= c * costs):
                theta = np.arcsin((TB-TA)/(c * costs))

                if ((np.maximum(0.0, alpha-0.5 * np.pi) <= theta) and (theta <= 0.5 * np.pi - beta)):
                    h = a * np.sin(alpha-theta)
                    TC = np.minimum(TC, h * costs + TB)
                else:
                    TC = np.minimum(np.minimum(TC, TA + b * costs), TB + a * costs)
            else:
                TC = np.minimum(np.minimum(TC, TA + b * costs), TB + a * costs)
            u.vec[p] = TC
 

def eikonal(gfphi, gfphiold, maxiter, mesh, vertex_to_elements, sort_index1, sort_index2, pointarray, costs):
    
    help = gfphi.vec.CreateVector()

    difference = 1e6
    iter = 1

    while (difference > 1e-6) and (iter <= maxiter):
        
        gfphiold.vec.data = gfphi.vec
        localsolver(gfphi, mesh, sort_index1, vertex_to_elements, pointarray, costs)
        localsolver(gfphi, mesh, sort_index1[::-1], vertex_to_elements, pointarray, costs)
        localsolver(gfphi, mesh, sort_index2, vertex_to_elements, pointarray, costs)
        localsolver(gfphi, mesh, sort_index2[::-1], vertex_to_elements, pointarray, costs)
        
        help.data = gfphi.vec - gfphiold.vec
        difference = InnerProduct(help, help)
        iter = iter + 1
